from __future__ import annotations

import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.template.backends.django import Template as DjangoTemplate
from django.template.backends.jinja2 import Template as JinjaTemplate
from django.test import override_settings
from model_bakery import baker

from django_simple_nav.nav import Nav
from django_simple_nav.nav import NavItem
from tests.navs import DummyNav
from tests.utils import count_anchors

pytestmark = pytest.mark.django_db


@pytest.mark.parametrize(
    "user, expected_count",
    [
        (AnonymousUser(), 7),
        (None, 10),
    ],
)
def test_nav_render(user, expected_count, req):
    if not isinstance(user, AnonymousUser):
        user = baker.make(get_user_model())

    req.user = user

    rendered_nav = DummyNav().render(req)

    assert count_anchors(rendered_nav) == expected_count


@pytest.mark.parametrize(
    "permission, expected_count",
    [
        ("", 10),  # regular authenticated user
        ("is_staff", 13),
        ("is_superuser", 19),
        ("tests.dummy_perm", 13),
    ],
)
def test_nav_render_permissions(req, permission, expected_count):
    user = baker.make(get_user_model())

    if permission == "tests.dummy_perm":
        dummy_perm = baker.make(
            "auth.Permission",
            codename="dummy_perm",
            name="Dummy Permission",
            content_type=baker.make("contenttypes.ContentType", app_label="tests"),
        )
        user.user_permissions.add(dummy_perm)
    else:
        setattr(user, permission, True)

    user.save()
    req.user = user

    rendered_nav = DummyNav().render(req)

    assert count_anchors(rendered_nav) == expected_count


def test_nav_render_template_name(req):
    req.user = AnonymousUser()

    rendered_nav = DummyNav().render(req, "tests/alternate.html")

    assert "This is an alternate template." in rendered_nav


def test_nav_render_template_string(req):
    class StringTemplateNav(Nav):
        title = ...
        items = [
            NavItem(title=..., url="/test/"),
        ]

        def get_template(self, template_name):
            return "<h1>This is a string.</h1>"

    rendered_nav = StringTemplateNav().render(req)

    assert "<h1>This is a string.</h1>" in rendered_nav
    assert count_anchors(rendered_nav) == 0


def test_get_context_data(req):
    context = DummyNav().get_context_data(req)

    assert context["items"] is not None


def test_get_context_data_override(req):
    class OverrideNav(DummyNav):
        def get_context_data(self, request):
            return {"foo": "bar"}

    context = OverrideNav().get_context_data(req)

    assert context["foo"] == "bar"


def test_get_items(req):
    class GetItemsNav(Nav):
        template_name = ...
        items = [
            NavItem(title=..., url=...),
        ]

    items = GetItemsNav().get_items(req)

    assert len(items) == 1


def test_get_items_override(req):
    class GetItemsNav(Nav):
        template_name = ...

        def get_items(self, request):
            return [
                NavItem(title=..., url=...),
            ]

    items = GetItemsNav().get_items(req)

    assert len(items) == 1


def test_get_items_improperly_configured(req):
    class GetItemsNav(Nav):
        template_name = ...

    with pytest.raises(ImproperlyConfigured):
        GetItemsNav().get_items(req)


@pytest.mark.parametrize(
    "engine,expected",
    [
        (
            "django.template.backends.django.DjangoTemplates",
            DjangoTemplate,
        ),
        (
            "django.template.backends.jinja2.Jinja2",
            JinjaTemplate,
        ),
    ],
)
def test_get_template_engines(engine, expected):
    class TemplateEngineNav(Nav):
        template_name = (
            "tests/dummy_nav.html"
            if engine.endswith("DjangoTemplates")
            else "tests/jinja2/dummy_nav.html"
        )
        items = [...]

    with override_settings(TEMPLATES=[dict(settings.TEMPLATES[0], BACKEND=engine)]):
        template = TemplateEngineNav().get_template()

    assert isinstance(template, expected)


def test_get_template_override(req):
    class TemplateOverrideNav(Nav):
        items = [NavItem(title=..., url="/test/")]

        def get_template(self, *args, **kwargs):
            return "<h1>Overridden Template</h1>"

    template = TemplateOverrideNav().get_template()

    assert isinstance(template, str)

    rendered_nav = TemplateOverrideNav().render(req)

    assert "<h1>Overridden Template</h1>" in rendered_nav


def test_get_template_argument():
    class TemplateOverrideNav(Nav):
        template_name = "foo.html"
        items = [...]

    template = TemplateOverrideNav().get_template(template_name="tests/dummy_nav.html")

    assert "tests/dummy_nav.html" in str(template.origin)
    assert "foo.html" not in str(template.origin)


def test_get_template_name():
    class GetTemplateNameNav(Nav):
        template_name = "tests/dummy_nav.html"
        items = [...]

    template_name = GetTemplateNameNav().get_template_name()

    assert template_name == "tests/dummy_nav.html"


def test_get_template_name_override():
    class GetTemplateNameNav(Nav):
        items = [...]

        def get_template_name(self):
            return "tests/dummy_nav.html"

    template_name = GetTemplateNameNav().get_template_name()

    assert template_name == "tests/dummy_nav.html"


def test_get_template_name_improperly_configured():
    class GetTemplateNameNav(Nav):
        items = [...]

    with pytest.raises(ImproperlyConfigured):
        GetTemplateNameNav().get_template_name()



from __future__ import annotations

import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import override_settings
from model_bakery import baker

from django_simple_nav.nav import NavGroup
from django_simple_nav.nav import NavItem

pytestmark = pytest.mark.django_db


def test_get_context_data(req):
    group = NavGroup(
        title="Test",
        items=[
            NavItem(
                title="Test",
                url="/test/",
                extra_context={"foo": "bar"},
            ),
        ],
        url="/test/",
        extra_context={"baz": "qux"},
    )

    group_context = group.get_context_data(req)

    assert {"title", "url", "active", "items", "baz"} == set(group_context.keys())
    assert group_context.get("title") == "Test"
    assert group_context.get("url") == "/test/"
    assert group_context.get("active") is False
    assert group_context.get("items") is not None
    assert len(group_context.get("items")) == 1
    assert group_context.get("baz") == "qux"

    item_context = group_context.get("items")[0]

    assert {"title", "url", "active", "items", "foo"} == set(item_context.keys())
    assert item_context.get("title") == "Test"
    assert item_context.get("url") == "/test/"
    assert item_context.get("active") is False
    assert item_context.get("items") is None
    assert item_context.get("foo") == "bar"


@pytest.mark.parametrize(
    "items,expected",
    [
        (
            [
                NavItem(title=..., url="/test/"),
                NavItem(title=..., url="/test2/"),
            ],
            2,
        ),
        (
            [
                NavItem(title=..., url="/test/", permissions=["is_authenticated"]),
                NavItem(title=..., url="/test2/"),
            ],
            1,
        ),
        (
            [
                NavItem(title=..., url="/test/", permissions=["is_authenticated"]),
                NavItem(title=..., url="/test2/", permissions=["is_authenticated"]),
            ],
            0,
        ),
    ],
)
def test_get_context_data_items_anonymous(items, expected, req):
    group = NavGroup(title=..., items=items)

    req.user = AnonymousUser()

    group_context = group.get_context_data(req)

    assert len(group_context.get("items")) == expected


@pytest.mark.parametrize(
    "items,expected",
    [
        (
            [
                NavItem(title=..., url="/test/"),
                NavItem(title=..., url="/test2/"),
            ],
            2,
        ),
        (
            [
                NavItem(title=..., url="/test/", permissions=["is_authenticated"]),
                NavItem(title=..., url="/test2/"),
            ],
            2,
        ),
        (
            [
                NavItem(title=..., url="/test/", permissions=["is_authenticated"]),
                NavItem(title=..., url="/test2/", permissions=["is_authenticated"]),
            ],
            2,
        ),
    ],
)
def test_get_context_data_items_is_authenticated(items, expected, req):
    group = NavGroup(title=..., items=items)

    req.user = baker.make(get_user_model())

    group_context = group.get_context_data(req)

    assert len(group_context.get("items")) == expected


def test_get_items(req):
    group = NavGroup(
        title=...,
        items=[
            NavItem(title=..., url="/test/"),
            NavItem(title=..., url="/test2/"),
        ],
    )

    items = group.get_items(req)

    assert items[0].url == "/test/"
    assert items[1].url == "/test2/"


def test_get_items_override(req):
    class GetItemsNavGroup(NavGroup):
        def get_items(self, request):
            items = super().get_items(request)
            return [
                NavItem(title=item.title, url=f"{item.url.rstrip('/')}/overridden/")
                for item in items
            ]

    group = GetItemsNavGroup(
        title=...,
        items=[
            NavItem(title=..., url="/test/"),
            NavItem(title=..., url="/test2/"),
        ],
    )

    items = group.get_items(req)

    assert items[0].url == "/test/overridden/"
    assert items[1].url == "/test2/overridden/"


@pytest.mark.parametrize(
    "url,append_slash,expected",
    [
        (None, True, ""),
        (None, False, ""),
        ("/test", True, "/test/"),
        ("/test", False, "/test"),
        ("home", True, "/"),
        ("home", False, "/"),
    ],
)
def test_get_url(url, append_slash, expected):
    group = NavGroup(title=..., url=url, items=[...])

    with override_settings(APPEND_SLASH=append_slash):
        rendered_url = group.get_url()

        if rendered_url not in ["", "/"]:
            assert rendered_url.endswith("/") is append_slash
        assert rendered_url == expected


def test_get_url_override():
    class GetURLNavGroup(NavGroup):
        def get_url(self):
            url = super().get_url()
            if url == "":
                raise ValueError
            return url

    group = GetURLNavGroup(title=..., items=[...])

    with pytest.raises(ValueError):
        group.get_url()


@pytest.mark.parametrize(
    "url,expected",
    [
        ("/test/", True),
        ("/test/active/", True),
        ("/foo/", False),
    ],
)
def test_get_active(url, expected, req):
    group = NavGroup(
        title=...,
        url="/test/",
        items=[
            NavItem(title=..., url="/test/active/"),
            NavItem(title=..., url="/test/not-active/"),
        ],
    )

    req.path = url

    assert group.get_active(req) is expected


@pytest.mark.parametrize(
    "url,items,expected",
    [
        (
            None,
            [
                NavItem(title=..., url=...),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=...),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            None,
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            None,
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
            ],
            False,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
            ],
            True,
        ),
    ],
)
def test_items_check_permissions_anonymous(url, items, expected, req):
    group = NavGroup(title=..., url=url, items=items)

    req.user = AnonymousUser()

    assert group.check_permissions(req) is expected


@pytest.mark.parametrize(
    "url,items,expected",
    [
        (
            None,
            [
                NavItem(title=..., url=...),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=...),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            None,
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=...),
            ],
            True,
        ),
        (
            None,
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
            ],
            True,
        ),
        (
            "/test/",
            [
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
                NavItem(title=..., url=..., permissions=["is_authenticated"]),
            ],
            True,
        ),
    ],
)
def test_items_check_permissions_is_authenticated(url, items, expected, req):
    group = NavGroup(title=..., url=url, items=items)

    req.user = baker.make(get_user_model())

    assert group.check_permissions(req) is expected



from __future__ import annotations

import pytest
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings

from django_simple_nav._templates import get_template_engine


def test_get_template_engine():
    engine = get_template_engine()

    assert engine.name in settings.TEMPLATES[0].get("BACKEND")


@override_settings(TEMPLATES=[])
def test_get_template_engine_no_engine():
    with pytest.raises(ImproperlyConfigured):
        get_template_engine()


@override_settings(
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
        },
        {
            "BACKEND": "django.template.backends.jinja2.Jinja2",
        },
    ]
)
@pytest.mark.parametrize(
    "templates,expected",
    [
        (
            [
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                },
                {
                    "BACKEND": "django.template.backends.jinja2.Jinja2",
                },
            ],
            "django",
        ),
        (
            [
                {
                    "BACKEND": "django.template.backends.jinja2.Jinja2",
                },
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                },
            ],
            "jinja2",
        ),
    ],
)
def test_get_template_engine_multiple(templates, expected, caplog):
    with override_settings(TEMPLATES=templates):
        with caplog.at_level("WARNING"):
            engine = get_template_engine()

    assert engine.name == expected
    assert "Multiple `BACKEND` defined for a template engine." in caplog.text


@pytest.mark.parametrize(
    "using,expected",
    [
        ("django", "django"),
        ("django2", "django2"),
        ("jinja2", "jinja2"),
    ],
)
@override_settings(
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
        },
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "NAME": "django2",
        },
        {
            "BACKEND": "django.template.backends.jinja2.Jinja2",
        },
    ]
)
def test_get_template_engine_using(using, expected):
    engine = get_template_engine(using)

    assert engine.name == expected


@pytest.mark.parametrize(
    "app_setting,expected",
    [
        ("django.template.backends.django.DjangoTemplates", "django"),
        ("django.template.backends.jinja2.Jinja2", "jinja2"),
    ],
)
def test_get_template_engine_app_setting(app_setting, expected):
    with override_settings(
        DJANGO_SIMPLE_NAV={"TEMPLATE_BACKEND": app_setting},
        TEMPLATES=[
            {
                "BACKEND": app_setting,
            },
        ],
    ):
        engine = get_template_engine()

    assert engine.name == expected


@override_settings(
    DJANGO_SIMPLE_NAV={"TEMPLATE_BACKEND": "invalid"},
)
def test_get_template_engine_app_setting_invalid():
    with pytest.raises(ImproperlyConfigured) as exc_info:
        get_template_engine()

        assert "Invalid `TEMPLATE_BACKEND` for a template engine" in exc_info



from __future__ import annotations

import logging

from django.core.exceptions import ImproperlyConfigured
from django.template import engines
from django.template.backends.base import BaseEngine

from django_simple_nav.conf import app_settings

logger = logging.getLogger(__name__)


def get_template_engine(using: str | None = None) -> BaseEngine:
    if template_backend := app_settings.TEMPLATE_BACKEND:
        # https://github.com/django/django/blob/082fe2b5a83571dec4aa97580af0fe8cf2a5214e/django/template/utils.py#L33-L42
        try:
            backend_alias = template_backend.rsplit(".", 2)[-2]
        except Exception as err:
            msg = f"Invalid `TEMPLATE_BACKEND` for a template engine: {app_settings.TEMPLATE_BACKEND}. Check your `DJANGO_SIMPLE_NAV['TEMPLATE_BACKEND']` setting."
            raise ImproperlyConfigured(msg) from err

        engine = engines[backend_alias]
    else:
        all_engines = engines.all()
        num_of_engines = len(all_engines)

        if num_of_engines == 0:
            msg = "No `BACKEND` found for a template engine. Please configure at least one in your `TEMPLATES` setting or set `DJANGO_SIMPLE_NAV['TEMPLATE_BACKEND']`."
            raise ImproperlyConfigured(msg)

        if num_of_engines > 1:
            msg = "Multiple `BACKEND` defined for a template engine. Will proceed with first defined in list, otherwise set `DJANGO_SIMPLE_NAV['TEMPLATE_BACKEND']` to specify which one to use."
            logger.warning(msg)

        # https://github.com/django/django/blob/082fe2b5a83571dec4aa97580af0fe8cf2a5214e/django/template/loader.py#L65-L66
        engine = all_engines[0] if using is None else engines[using]

    return engine



from __future__ import annotations

import logging
from dataclasses import dataclass
from dataclasses import field
from typing import Callable
from typing import cast
from urllib.parse import parse_qs
from urllib.parse import urlparse
from urllib.parse import urlunparse

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest
from django.template.loader import get_template
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.functional import Promise
from django.utils.safestring import mark_safe

from ._templates import get_template_engine
from ._typing import EngineTemplate
from ._typing import override

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Nav:
    template_name: str | None = field(init=False, default=None)
    items: list[NavGroup | NavItem] | None = field(init=False, default=None)

    def render(self, request: HttpRequest, template_name: str | None = None) -> str:
        context = self.get_context_data(request)
        template = self.get_template(template_name)
        if isinstance(template, str):
            engine = get_template_engine()
            template = engine.from_string(template)
        return template.render(context, request)

    def get_context_data(self, request: HttpRequest) -> dict[str, object]:
        items = self.get_items(request)
        return {
            "items": [item.get_context_data(request) for item in items],
        }

    def get_items(self, request: HttpRequest) -> list[NavGroup | NavItem]:
        if self.items is not None:
            return [item for item in self.items if item.check_permissions(request)]

        msg = f"{self.__class__!r} must define 'items' or override 'get_items()'"
        raise ImproperlyConfigured(msg)

    def get_template(self, template_name: str | None = None) -> EngineTemplate | str:
        return get_template(template_name=template_name or self.get_template_name())

    def get_template_name(self) -> str:
        if self.template_name is not None:
            return self.template_name

        msg = f"{self.__class__!r} must define 'template_name' or override 'get_template_name()'"
        raise ImproperlyConfigured(msg)


@dataclass(frozen=True)
class NavItem:
    title: str
    url: str | Callable[..., str] | Promise | None = None
    permissions: list[str | Callable[[HttpRequest], bool]] = field(default_factory=list)
    extra_context: dict[str, object] = field(default_factory=dict)

    def get_context_data(self, request: HttpRequest) -> dict[str, object]:
        context = {
            "title": self.get_title(),
            "url": self.get_url(),
            "active": self.get_active(request),
            "items": self.get_items(request),
        }
        # filter out any items in `extra_context` that may be shadowing the
        # above `context` dict
        extra_context = {
            key: value
            for key, value in self.extra_context.copy().items()
            if context.get(key) is None
        }
        return {
            **context,
            **extra_context,
        }

    def get_title(self) -> str:
        return mark_safe(self.title)

    def get_url(self) -> str:
        url: str | None

        if isinstance(self.url, Promise):
            # django.urls.base.reverse_lazy
            url = str(self.url)
        elif callable(self.url):
            # django.urls.base.reverse (or some other basic callable)
            url = self.url()
        else:
            try:
                url = reverse(self.url)
            except NoReverseMatch:
                url = self.url

        if url is not None:
            parsed_url = urlparse(url)
            path = parsed_url.path
            if settings.APPEND_SLASH and not path.endswith("/"):
                path += "/"
            url = urlunparse(
                (
                    parsed_url.scheme,
                    parsed_url.netloc,
                    path,
                    parsed_url.params,
                    parsed_url.query,
                    parsed_url.fragment,
                )
            )
            return url

        msg = f"{self.__class__!r} must define 'url' or override 'get_url()'"
        raise ImproperlyConfigured(msg)

    def get_active(self, request: HttpRequest) -> bool:
        try:
            url = self.get_url()
        except ImproperlyConfigured:
            url = None

        if not url:
            return False

        parsed_url = urlparse(url)
        parsed_request = urlparse(request.build_absolute_uri())

        url_path = parsed_url.path
        request_path = parsed_request.path

        if settings.APPEND_SLASH:
            url_path = url_path.rstrip("/") + "/"
            request_path = request_path.rstrip("/") + "/"

        url_query = parse_qs(parsed_url.query)
        request_query = parse_qs(parsed_request.query)

        return url_path == request_path and url_query == request_query

    def get_items(self, request: HttpRequest) -> list[NavGroup | NavItem] | None:
        # this needs to be set to shadow the built-in `items()` of the dict
        # returned by this method for `NavItem`. otherwise when looping through
        # the items in a `Nav`, calling `{% if item.items %}` will resolve to `True`.
        # this is a consequence of getting rid of `RenderedNavItem` from earlier versions
        # of `django-simple-nav` which was done to slightly simplfy the library. though
        # given this hack, maybe that wasn't really worth it? idk... i guess it could
        # have been called `children` or `subnav`, but i started with `items` and
        # i'd rather have API consistency and this hack.
        return None

    def check_permissions(self, request: HttpRequest) -> bool:
        if not apps.is_installed("django.contrib.auth"):
            logger.warning(
                "The 'django.contrib.auth' app is not installed, so permissions will not be checked."
            )
            return True

        if not hasattr(request, "user"):
            # if no user attached to request, we assume that the user is not authenticated
            # and we should hide if *any* permissions are set
            return not self.permissions

        if not self.permissions:
            return True

        # explicitly cast to AbstractUser to make static type checkers happy
        # `django-stubs` types `request.user` as `django.contrib.auth.base_user.AbstractBaseUser`
        # as opposed to `django.contrib.auth.models.AbstractUser` or `django.contrib.auth.models.User`
        # so any type checkers will complain if this is not casted
        user = cast(AbstractUser, request.user)

        permission_checks: list[bool] = []

        for idx, perm in enumerate(self.permissions):
            has_perm = False

            if getattr(user, "is_superuser", False):
                permission_checks.append(True)
                break
            elif callable(perm):
                has_perm = perm(request)
            elif perm in ["is_authenticated", "is_staff"]:
                has_perm = getattr(user, perm, False)
            else:
                has_perm = user.has_perm(perm)

            permission_checks.append(has_perm)

            if not idx == len(self.permissions) - 1:
                continue

        return all(permission_checks)


@dataclass(frozen=True)
class NavGroup(NavItem):
    items: list[NavGroup | NavItem] = field(default_factory=list)

    @override
    def get_context_data(self, request: HttpRequest) -> dict[str, object]:
        context = super().get_context_data(request)

        items = self.get_items(request)
        context["items"] = [item.get_context_data(request) for item in items]

        return context

    @override
    def get_items(self, request: HttpRequest) -> list[NavGroup | NavItem]:
        return [item for item in self.items if item.check_permissions(request)]

    @override
    def get_url(self) -> str:
        try:
            url = super().get_url()
        except ImproperlyConfigured:
            return ""
        return url

    @override
    def get_active(self, request: HttpRequest) -> bool:
        is_active = super().get_active(request)

        items = self.get_items(request)
        item_is_active = any([item.get_active(request) for item in items])

        return is_active or item_is_active

    @override
    def check_permissions(self, request: HttpRequest) -> bool:
        has_perm = super().check_permissions(request)

        sub_items = [
            sub_item
            for sub_item in self.items
            if sub_item.check_permissions(request) is not False
        ]
        if not sub_items and not self.url:
            has_perm = False

        return has_perm



from __future__ import annotations

from dataclasses import dataclass

from django.conf import settings

from ._typing import override

DJANGO_SIMPLE_NAV_SETTINGS_NAME = "DJANGO_SIMPLE_NAV"


@dataclass(frozen=True)
class AppSettings:
    TEMPLATE_BACKEND: str | None = None

    @override
    def __getattribute__(self, __name: str) -> object:
        user_settings = getattr(settings, DJANGO_SIMPLE_NAV_SETTINGS_NAME, {})
        return user_settings.get(__name, super().__getattribute__(__name))  # pyright: ignore[reportAny]


app_settings = AppSettings()



from __future__ import annotations

import sys
from typing import Protocol

from django.http import HttpRequest
from django.template.context import Context
from django.utils.safestring import SafeString

if sys.version_info >= (3, 12):
    from typing import override as typing_override
else:  # pragma: no cover
    from typing_extensions import (
        override as typing_override,  # pyright: ignore[reportUnreachable]
    )

override = typing_override


# https://github.com/typeddjango/django-stubs/blob/b6e8ea9b4279ece87d14e38226c265e4f0aadccd/django-stubs/template/backends/base.pyi#L22-L28
class EngineTemplate(Protocol):
    def render(
        self,
        context: Context | dict[str, object] | None = ...,
        request: HttpRequest | None = ...,
    ) -> SafeString: ...



from __future__ import annotations

from django.urls import path


def home(request): ...


def fake_view(request): ...


urlpatterns = [
    path("fake-view/", fake_view, name="fake-view"),
    path("", home, name="home"),
]



from __future__ import annotations

import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings
from django.urls import reverse
from django.urls import reverse_lazy
from model_bakery import baker

from django_simple_nav.nav import NavItem

pytestmark = pytest.mark.django_db


def test_get_context_data(req):
    item = NavItem(
        title="Test",
        url="/test/",
        permissions=["is_authenticated"],
        extra_context={"foo": "bar"},
    )

    context = item.get_context_data(req)

    assert {"title", "url", "active", "items", "foo"} == set(context.keys())
    assert context.get("title") == "Test"
    assert context.get("url") == "/test/"
    assert context.get("active") is False
    assert context.get("items") is None
    assert context.get("foo") == "bar"


def test_get_context_data_no_extra_context(req):
    item = NavItem(
        title="Test",
        url="/test/",
    )

    rendered_item = item.get_context_data(req)

    assert rendered_item.get("foo") is None


def test_get_context_data_extra_context_shadowing(req):
    item = NavItem(
        title="Test",
        url="/test/",
        extra_context={"title": "Shadowed"},
    )

    rendered_item = item.get_context_data(req)

    assert rendered_item.get("title") == "Test"


def test_get_title():
    item = NavItem(title="Test")

    assert item.get_title() == "Test"


def test_get_title_safestring():
    item = NavItem(title="Test!!!")

    assert item.get_title() == "Test!!!"


def test_get_title_override():
    class GetTitleNavItem(NavItem):
        def get_title(self):
            return f"{self.title}!!!"

    item = GetTitleNavItem(title="Test")

    assert item.get_title() == "Test!!!"


@pytest.mark.parametrize(
    "url,append_slash,expected",
    [
        ("/test", True, "/test/"),
        ("/test", False, "/test"),
        ("home", True, "/"),
        ("home", False, "/"),
        (reverse("home"), True, "/"),
        (reverse("home"), False, "/"),
        (reverse_lazy("home"), True, "/"),
        (reverse_lazy("home"), False, "/"),
    ],
)
def test_get_url(url, append_slash, expected):
    item = NavItem(title=..., url=url)

    with override_settings(APPEND_SLASH=append_slash):
        rendered_url = item.get_url()

        if rendered_url != "/":
            assert rendered_url.endswith("/") is append_slash
        assert rendered_url == expected


def test_get_url_improperly_configured():
    item = NavItem(title=..., url=None)

    with pytest.raises(ImproperlyConfigured):
        item.get_url()


def test_get_url_override():
    class GetURLNavItem(NavItem):
        def get_url(self):
            return "/"

    item = GetURLNavItem(title=..., url=None)

    assert item.get_url() == "/"


@pytest.mark.parametrize(
    "url,req_path,req_params,expected",
    [
        ("/test/", "/test/", None, True),
        ("/test/", "/other/", None, False),
        ("fake-view", "/fake-view/", None, True),
        ("/test", "/test/", None, True),
        ("/test/", "/test", None, True),
        ("/test/nested/", "/test/", None, False),
        ("/test/?query=param", "/test/", {"query": "param"}, True),
        ("/test/?query=param", "/test/", None, False),
    ],
)
def test_active(url, req_path, req_params, expected, rf):
    item = NavItem(title=..., url=url)

    req = rf.get(req_path, req_params)
    print(f"{req.path=}")

    assert item.get_active(req) == expected


@pytest.mark.parametrize("append_slash", [True, False])
def test_active_append_slash_setting(append_slash, req):
    item = NavItem(title=..., url="/test")

    req.path = "/test"

    with override_settings(APPEND_SLASH=append_slash):
        assert item.get_active(req) is True


def test_active_improperly_configured(req):
    item = NavItem(title=..., url=None)

    req.path = "/"

    assert item.get_active(req) is False


def test_active_reverse_no_match(req):
    item = NavItem(title=..., url="nonexistent")

    req.path = "/"

    assert item.get_active(req) is False


def test_get_items(req):
    item = NavItem(title=..., url=...)

    assert item.get_items(req) is None


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], False),
        (["is_staff"], False),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], False),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_anonymous(permissions, expected, req):
    item = NavItem(title=..., url=..., permissions=permissions)

    req.user = AnonymousUser()

    assert item.check_permissions(req) is expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], False),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], False),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_is_authenticated(permissions, expected, req):
    item = NavItem(title=..., url=..., permissions=permissions)

    req.user = baker.make(get_user_model())

    assert item.check_permissions(req) is expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], True),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], True),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_is_staff(permissions, expected, req):
    item = NavItem(title=..., url=..., permissions=permissions)

    req.user = baker.make(get_user_model(), is_staff=True)

    assert item.check_permissions(req) is expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], True),
        (["is_superuser"], True),
        (["is_authenticated", "is_staff"], True),
        (["is_authenticated", "is_superuser"], True),
    ],
)
def test_check_permissions_is_superuser(permissions, expected, req):
    item = NavItem(title=..., url=..., permissions=permissions)

    req.user = baker.make(get_user_model(), is_superuser=True)

    assert item.check_permissions(req) is expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], False),
        (["is_staff"], False),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], False),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_no_request_user(permissions, expected, req):
    item = NavItem(title=..., url=..., permissions=permissions)

    assert item.check_permissions(req) is expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], True),
        (["is_superuser"], True),
        (["is_authenticated", "is_staff"], True),
        (["is_authenticated", "is_superuser"], True),
    ],
)
@override_settings(
    INSTALLED_APPS=[
        app for app in settings.INSTALLED_APPS if app != "django.contrib.auth"
    ]
)
def test_check_permissions_no_contrib_auth(permissions, expected, req, caplog):
    item = NavItem(title=..., url=..., permissions=permissions)

    with caplog.at_level("WARNING"):
        assert item.check_permissions(req) is expected

    assert "The 'django.contrib.auth' app is not installed" in caplog.text


def test_check_permissions_callable_anonymous(req):
    def dummy_check(request):
        return True

    item = NavItem(title=..., url=..., permissions=[dummy_check])

    req.user = AnonymousUser()

    assert item.check_permissions(req)


@pytest.mark.parametrize("is_authenticated", [True, False])
def test_check_permissions_callable_is_authenticated(is_authenticated, req):
    def check_is_authenticated(request):
        return request.user.is_authenticated

    item = NavItem(title=..., url=..., permissions=[check_is_authenticated])

    req.user = baker.make(get_user_model()) if is_authenticated else AnonymousUser()

    assert item.check_permissions(req) is is_authenticated


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], False),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], False),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_auth_permission_is_authenticated(permissions, expected, req):
    dummy_perm = baker.make(
        "auth.Permission",
        codename="dummy_perm",
        name="Dummy Permission",
        content_type=baker.make("contenttypes.ContentType", app_label="tests"),
    )

    item = NavItem(
        title=...,
        url=...,
        permissions=permissions
        + [f"{dummy_perm.content_type.app_label}.{dummy_perm.codename}"],
    )

    user = baker.make(get_user_model())
    user.user_permissions.add(dummy_perm)
    req.user = user

    assert item.check_permissions(req) == expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], True),
        (["is_superuser"], False),
        (["is_authenticated", "is_staff"], True),
        (["is_authenticated", "is_superuser"], False),
    ],
)
def test_check_permissions_auth_permission_is_staff(permissions, expected, req):
    dummy_perm = baker.make(
        "auth.Permission",
        codename="dummy_perm",
        name="Dummy Permission",
        content_type=baker.make("contenttypes.ContentType", app_label="tests"),
    )

    item = NavItem(
        title=...,
        url=...,
        permissions=permissions
        + [f"{dummy_perm.content_type.app_label}.{dummy_perm.codename}"],
    )

    user = baker.make(get_user_model(), is_staff=True)
    user.user_permissions.add(dummy_perm)
    req.user = user

    assert item.check_permissions(req) == expected


@pytest.mark.parametrize(
    "permissions,expected",
    [
        ([], True),
        (["is_authenticated"], True),
        (["is_staff"], True),
        (["is_superuser"], True),
        (["is_authenticated", "is_staff"], True),
        (["is_authenticated", "is_superuser"], True),
    ],
)
def test_check_permissions_auth_permission_is_superuser(permissions, expected, req):
    dummy_perm = baker.make(
        "auth.Permission",
        codename="dummy_perm",
        name="Dummy Permission",
        content_type=baker.make("contenttypes.ContentType", app_label="tests"),
    )

    item = NavItem(
        title=...,
        url=...,
        permissions=permissions
        + [f"{dummy_perm.content_type.app_label}.{dummy_perm.codename}"],
    )

    user = baker.make(get_user_model(), is_superuser=True)
    user.user_permissions.add(dummy_perm)
    req.user = user

    assert item.check_permissions(req) == expected



from __future__ import annotations

from django import template
from django.http import HttpRequest
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context
from django.utils.module_loading import import_string

from django_simple_nav._typing import override
from django_simple_nav.nav import Nav

register = template.Library()


@register.tag(name="django_simple_nav")
def do_django_simple_nav(parser: Parser, token: Token) -> DjangoSimpleNavNode:
    try:
        args = token.split_contents()[1:]
        if len(args) == 0:
            raise ValueError
    except ValueError as err:
        raise template.TemplateSyntaxError(
            f"{token.split_contents()[0]} tag requires arguments"
        ) from err

    nav = args[0]
    template_name = args[1] if len(args) > 1 else None

    return DjangoSimpleNavNode(nav, template_name)


class DjangoSimpleNavNode(template.Node):
    def __init__(self, nav: str, template_name: str | None) -> None:
        self.nav = template.Variable(nav)
        self.template_name = template.Variable(template_name) if template_name else None

    @override
    def render(self, context: Context) -> str:
        nav = self.get_nav(context)
        template_name = self.get_template_name(context)
        request = self.get_request(context)

        return nav.render(request, template_name)

    def get_nav(self, context: Context) -> Nav:
        try:
            nav: str | Nav = self.nav.resolve(context)
        except template.VariableDoesNotExist as err:
            raise template.TemplateSyntaxError(
                f"Variable does not exist: {err}"
            ) from err

        if isinstance(nav, str):
            try:
                nav_instance: object = import_string(nav)()
            except ImportError as err:
                raise template.TemplateSyntaxError(
                    f"Failed to import from dotted string: {nav}"
                ) from err
        else:
            nav_instance = nav

        if not isinstance(nav_instance, Nav):
            raise template.TemplateSyntaxError(
                f"Not a valid `Nav` instance: {nav_instance}"
            )

        return nav_instance

    def get_template_name(self, context: Context) -> str | None:
        try:
            template_name = (
                self.template_name.resolve(context) if self.template_name else None
            )
        except template.VariableDoesNotExist as err:
            raise template.TemplateSyntaxError(
                f"Variable does not exist: {err}"
            ) from err

        return template_name

    def get_request(self, context: Context) -> HttpRequest:
        request = context.get("request", None)

        if not request:
            raise template.TemplateSyntaxError(
                f"`request` not found in template context: {context}"
            )

        if not isinstance(request, HttpRequest):
            raise template.TemplateSyntaxError(
                f"`request` not a valid `HttpRequest`: {request}"
            )

        return request
